@@ -35,7 +35,8 @@
 
 int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 	  int raiddisks,
-	  mddev_dev_t devlist, int assume_clean)
+	  mddev_dev_t devlist, int assume_clean,
+	  char *bitmap_file, int bitmap_chunk, int write_behind, int delay, int verbose)
 {
 	/* Build a linear or raid0 arrays without superblocks
 	 * We cannot really do any checks, we just do it.
@@ -50,15 +51,21 @@ int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 	 * SET_ARRAY_INFO,  ADD_NEW_DISK, RUN_ARRAY
 	 *
 	 */
-	int verbose = 0;
 	int i;
 	int vers;
 	struct stat stb;
-	int subdevs = 0;
+	int subdevs = 0, missing_disks = 0;
 	mddev_dev_t dv;
+	int bitmap_fd;
+	unsigned long long size = ~0ULL;
 
 	/* scan all devices, make sure they really are block devices */
 	for (dv = devlist; dv; dv=dv->next) {
+		subdevs++;
+		if (strcmp("missing", dv->devname) == 0) {
+			missing_disks++;
+			continue;
+		}
 		if (stat(dv->devname, &stb)) {
 			fprintf(stderr, Name ": Cannot find %s: %s\n",
 				dv->devname, strerror(errno));
@@ -69,7 +76,6 @@ int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 				dv->devname);
 			return 1;
 		}
-		subdevs++;
 	}
 
 	if (raiddisks != subdevs) {
@@ -85,21 +91,21 @@ int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 			break;
 		case 10:
 			layout = 0x102; /* near=2, far=1 */
-			if (verbose)
+			if (verbose > 0)
 				fprintf(stderr,
 					Name ": layout defaults to n1\n");
 			break;
 		case 5:
 		case 6:
 			layout = map_name(r5layout, "default");
-			if (verbose)
+			if (verbose > 0)
 				fprintf(stderr,
 					Name ": layout defaults to %s\n", map_num(r5layout, layout));
 			break;
 		case LEVEL_FAULTY:
 			layout = map_name(faultylayout, "default");
 
-			if (verbose)
+			if (verbose > 0)
 				fprintf(stderr,
 					Name ": layout defaults to %s\n", map_num(faultylayout, layout));
 			break;
@@ -122,11 +128,11 @@ int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 		array.state = 0; /* not clean, but no errors */
 		if (assume_clean)
 			array.state |= 1;
-		array.active_disks = raiddisks;
-		array.working_disks = raiddisks;
+		array.active_disks = raiddisks - missing_disks;
+		array.working_disks = raiddisks - missing_disks;
 		array.spare_disks = 0;
-		array.failed_disks = 0;
-		if (chunk == 0)  
+		array.failed_disks = missing_disks;
+		if (chunk == 0 && (level==0 || level==LEVEL_LINEAR))
 			chunk = 64;
 		array.chunk_size = chunk*1024;
 		array.layout = layout;
@@ -135,11 +141,18 @@ int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 				mddev, strerror(errno));
 			return 1;
 		}
+	} else if (bitmap_file) {
+		fprintf(stderr, Name ": bitmaps not supported with this kernel\n");
+		return 1;
 	}
 	/* now add the devices */
 	for ((i=0), (dv = devlist) ; dv ; i++, dv=dv->next) {
+		unsigned long dsize;
+		int fd;
+		if (strcmp("missing", dv->devname) == 0)
+			continue;
 		if (stat(dv->devname, &stb)) {
-			fprintf(stderr, Name ": Wierd: %s has disappeared.\n",
+			fprintf(stderr, Name ": Weird: %s has disappeared.\n",
 				dv->devname);
 			goto abort;
 		}
@@ -148,11 +161,26 @@ int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 				dv->devname);
 			goto abort;
 		}
+		fd = open(dv->devname, O_RDONLY|O_EXCL);
+		if (fd < 0) {
+			fprintf(stderr, Name ": Cannot open %s: %s\n", 
+				dv->devname, strerror(errno));
+			goto abort;
+		}
+		if (ioctl(fd, BLKGETSIZE, &dsize) == 0 && dsize > 0) {
+			unsigned long long ldsize = dsize;
+			ldsize <<= 9;
+			if (size== 0 || ldsize < size)
+				size = ldsize;
+		}
+		close(fd);
 		if (vers>= 9000) {
 			mdu_disk_info_t disk;
 			disk.number = i;
 			disk.raid_disk = i;
-			disk.state = 6;
+			disk.state = (1<<MD_DISK_SYNC) | (1<<MD_DISK_ACTIVE);
+			if (dv->writemostly)
+				disk.state |= 1<<MD_DISK_WRITEMOSTLY;
 			disk.major = major(stb.st_rdev);
 			disk.minor = minor(stb.st_rdev);
 			if (ioctl(mdfd, ADD_NEW_DISK, &disk)) {
@@ -171,6 +199,33 @@ int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 	/* now to start it */
 	if (vers >= 9000) {
 		mdu_param_t param; /* not used by syscall */
+		if (bitmap_file) {
+			bitmap_fd = open(bitmap_file, O_RDWR);
+			if (bitmap_fd < 0) {
+				if (bitmap_chunk == UnSet) {
+					fprintf(stderr, Name ": %s cannot be openned.",
+						bitmap_file);
+					return 1;
+				}
+				if (CreateBitmap(bitmap_file, 1, NULL, bitmap_chunk,
+						 delay, write_behind, size>>9)) {
+					return 1;
+				}
+				bitmap_fd = open(bitmap_file, O_RDWR);
+				if (bitmap_fd < 0) {
+					fprintf(stderr, Name ": %s cannot be openned.",
+						bitmap_file);
+					return 1;
+				}
+			}				
+			if (bitmap_fd >= 0) {
+				if (ioctl(mdfd, SET_BITMAP_FILE, bitmap_fd) < 0) {
+					fprintf(stderr, Name ": Cannot set bitmap file for %s: %s\n",
+						mddev, strerror(errno));
+					return 1;
+				}
+			}
+		}
 		if (ioctl(mdfd, RUN_ARRAY, &param)) {
 			fprintf(stderr, Name ": RUN_ARRAY failed: %s\n",
 				strerror(errno));
@@ -192,8 +247,9 @@ int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 			goto abort;
 		}
 	}
-	fprintf(stderr, Name ": array %s built and started.\n",
-		mddev);
+	if (verbose >= 0)
+		fprintf(stderr, Name ": array %s built and started.\n",
+			mddev);
 	return 0;
 
  abort:
@@ -204,4 +260,3 @@ int Build(char *mddev, int mdfd, int chunk, int level, int layout,
 	return 1;
 		
 }
-
