@@ -198,7 +198,7 @@ bitmap_info_t *bitmap_file_read(char *filename, int brief, struct supertype **st
 			/* just look at device... */
 			lseek(fd, 0, 0);
 		} else {	
-			st->ss->locate_bitmap(st, fd);
+			st->ss->locate_bitmap(st, fd, NULL);
 		}
 		ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */
 		*stp = st;
